//============================================================
//  TARGET PROJECT: *
//  TARGET SYSTEM : *
//  FUNCTION      : Simple Uart
//
//                      Copyright (C) 2008 by Tomoyuki Inomoto
//============================================================

//=======================================================
//   UART TOP
//=======================================================
module Uart(
    Clk         ,
    RstN        ,
    UartCKx16   ,

    RxD         ,
    TxD         ,

    RComp       ,
    RDataRe     ,
    RData       ,

    SEmpty      ,
    SDataWe     ,
    SData
);
    //==================================================
    //======== port  ===================================
    //==================================================
    //======== SYSTEM ==============
    input           Clk         ;
    input           RstN        ;
    input           UartCKx16   ;   // Recv & Send Data Rate x 16

    //======== UART ==============
    input           RxD         ;   // UART RxD
    output          TxD         ;   // UART TxD

    //======== RECV DATA I/F ==============
    output          RComp       ;   // Recv Complete
    input           RDataRe     ;   // Recv Data Read Enable
    output  [7:0]   RData       ;   // Recv Data

    //======== SEND DATA I/F ==============
    output          SEmpty      ;   // Send Buf Empty
    input           SDataWe     ;   // Send Data Write Enable
    input   [7:0]   SData       ;   // Send Data

    //==================================================
    //======== wire & reg ==============================
    //==================================================

    //====================================================================
    //======== BODY ======================================================
    //====================================================================
    //===== Recv ===========================
    UartRecv Recv(
        .Clk        ( Clk       ),
        .RstN       ( RstN      ),
        .UartCKx16  ( UartCKx16 ),
        .RxD        ( RxD       ),
        .RComp      ( RComp     ),
        .RDataRe    ( RDataRe   ),
        .RData      ( RData     )
    );

    //===== Send ===========================
    UartSend Send(
        .Clk        ( Clk       ),
        .RstN       ( RstN      ),
        .UartCKx16  ( UartCKx16 ),
        .TxD        ( TxD       ),
        .SEmpty     ( SEmpty    ),
        .SDataWe    ( SDataWe   ),
        .SData      ( SData     )
    );

endmodule

//===============================================================
//   Recv
//===============================================================
`define SU_RECV_IDLE      4'h0
`define SU_RECV_START     4'h1
`define SU_RECV_BIT0      4'h2
`define SU_RECV_BIT1      4'h3
`define SU_RECV_BIT2      4'h4
`define SU_RECV_BIT3      4'h5
`define SU_RECV_BIT4      4'h6
`define SU_RECV_BIT5      4'h7
`define SU_RECV_BIT6      4'h8
`define SU_RECV_BIT7      4'h9
`define SU_RECV_STOP      4'hA

module UartRecv(
    Clk         ,
    RstN        ,
    UartCKx16   ,
    RxD         ,
    RComp       ,
    RDataRe     ,
    RData
);
    //==================================================
    //======== port  ===================================
    //==================================================
    //======== SYSTEM ==============
    input           Clk         ;
    input           RstN        ;

    input           UartCKx16   ;   // Recv Data Rate x 16

    //======== UART ==============
    input           RxD         ;   // UART RxD

    //======== RECV DATA I/F ==============
    output          RComp       ;   // Recv Complete
    input           RDataRe     ;   // Recv Data Read Enable
    output  [7:0]   RData       ;   // Recv Data


    //==================================================
    //======== wire & reg ==============================
    //==================================================
    reg             RComp;
    reg     [7:0]   RData;
    reg     [3:0]   RecvSampCount;
    reg             RecvTimClk;
    reg     [1:0]   RecvTimClkTmp;
    reg     [3:0]   RecvState;
    reg             DetectStart;
    reg     [1:0]   RxInput;
    reg     [7:0]   RecvBuf;

    //====================================================================
    //======== BODY ======================================================
    //====================================================================
    //====== RecvSampCount ===========
    always @(posedge Clk or negedge RstN) begin
        if(~RstN)                           RecvSampCount <= 4'h0;
        else if(RecvState == `SU_RECV_IDLE) RecvSampCount <= 4'h0;
        else if(UartCKx16)                  RecvSampCount <= RecvSampCount + 1'b1;
    end


    //====== RecvTimingClk ===========
    always @(posedge Clk) RecvTimClkTmp <= {RecvTimClkTmp[0], (RecvSampCount == 4'h8)};
    always @(posedge Clk) RecvTimClk    <= (RecvTimClkTmp == 2'b01);

    //=============================================================
    //===== RecvStateMachine ======================================
    //=============================================================
    //====== DetectStart ===========
    always @(posedge Clk) RxInput       <= {RxInput[0], RxD};
    always @(posedge Clk) DetectStart   <= (RecvState == `SU_RECV_IDLE) & (RxInput == 2'b10);

    //====== RecvStateMachine ===========
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
            RecvState <= `SU_RECV_IDLE;
        end
        else if(RecvState == `SU_RECV_IDLE) begin
            if(DetectStart) begin
                RecvState <= RecvState + 1'b1;
            end
        end
        else if(RecvTimClk) begin
            if(RecvState == `SU_RECV_STOP) begin
                RecvState <= `SU_RECV_IDLE;
            end
            else begin
                RecvState <= RecvState + 1'b1;
            end
        end
    end

    //======- RecvDataBuf ============
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
            RecvBuf <= 8'h0;
        end
        else if((RecvState > `SU_RECV_START) & (RecvState < `SU_RECV_STOP)) begin
            if(RecvTimClk) begin
                RecvBuf <= {RxInput[0], RecvBuf[7:1]};
            end
        end
    end
    //======- RData ============
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
            RData <= 8'h0;
        end
        else if((RecvState == `SU_RECV_STOP) & RecvTimClk) begin
            RData <= RecvBuf;
        end
    end

    //======- RComp ==========
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
            RComp <= 1'b0;
        end
        else begin
            if((RecvState == `SU_RECV_STOP) & RecvTimClk) begin
                RComp <= 1'b1;
            end
            else if(RDataRe) begin
                RComp <= 1'b0;
            end
        end
    end

endmodule
//===============================================================
//   Send
//===============================================================
`define SU_SEND_IDLE      4'h0
`define SU_SEND_START     4'h1
`define SU_SEND_BIT7      4'h2
`define SU_SEND_BIT6      4'h3
`define SU_SEND_BIT5      4'h4
`define SU_SEND_BIT4      4'h5
`define SU_SEND_BIT3      4'h6
`define SU_SEND_BIT2      4'h7
`define SU_SEND_BIT1      4'h8
`define SU_SEND_BIT0      4'h9
`define SU_SEND_STOP      4'hA
`define SU_SEND_STOP2     4'hB

module UartSend(
    Clk,
    RstN,
    UartCKx16,
    TxD,
    SEmpty,
    SDataWe,
    SData
);
    //==================================================
    //======== port  ===================================
    //==================================================
    //======== SYSTEM ==============
    input           Clk         ;
    input           RstN        ;
    input           UartCKx16   ;   // Send Data Rate x 16

    //======== UART ==============
    output          TxD         ;   // UART TxD

    //======== SEND DATA I/F ==============
    output          SEmpty      ;   // Send Buf Empty
    input           SDataWe     ;   // Send Data Write Enable
    input   [7:0]   SData       ;   // Send Data

    //==================================================
    //======== wire & reg ==============================
    //==================================================
    reg             TxD             ;
    reg             SEmpty          ;
    reg     [3:0]   SendSampCount   ;
    reg             SendTimClk      ;
    reg     [1:0]   SendTimClkTmp   ;
    reg     [3:0]   SendState       ;
    reg     [7:0]   SendBuf         ;

    //====================================================================
    //======== BODY ======================================================
    //====================================================================
    //====== SendSampCount ===========
    always @(posedge Clk or negedge RstN) begin
        if(~RstN)                           SendSampCount <= 4'h0;
        else if(SendState == `SU_SEND_IDLE) SendSampCount <= 4'h0;
        else if(UartCKx16)                  SendSampCount <= SendSampCount + 1'b1;
    end
    //====== SendTimingClk ===========
    always @(posedge Clk) SendTimClkTmp <= {SendTimClkTmp[0], (SendSampCount == 4'h7)};
    always @(posedge Clk) SendTimClk    <= (SendTimClkTmp == 2'b01);

    //======================================================
    //===== SendStateMachine ===============================
    //======================================================
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
            SendState <= `SU_SEND_IDLE;
        end
        else if(SendState == `SU_SEND_IDLE) begin
            if(SDataWe) begin
                SendState <= SendState + 1'b1;
            end
        end
        else if(SendTimClk) begin
            if(SendState == `SU_SEND_STOP2) begin
                SendState <= `SU_SEND_IDLE;
            end
            else begin
                SendState <= SendState + 1'b1;
            end
        end
    end

    //====== SendBuf ============
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
            SendBuf <= 8'hFF;
        end
        else if(SDataWe) begin
            SendBuf <= SData;
        end
        else if((SendState > `SU_SEND_START) & (SendState < `SU_SEND_STOP)) begin
            if(SendTimClk) begin
                SendBuf <= {1'b1, SendBuf[7:1]};
            end
        end
    end

    //====== TxD ============
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
            TxD <= 1'h1;
        end
        else if(SendTimClk) begin
            if(SendState == `SU_SEND_START) begin
                TxD <= 1'h0;
            end
            else begin
                TxD <= SendBuf[0];
            end
        end
   end

    //======- Send Complete =======
    always @(posedge Clk or negedge RstN) begin
        if(~RstN)   SEmpty <= 1'b1;
        else begin
            if((SendState == `SU_SEND_STOP2) & SendTimClk) begin
                SEmpty <= 1'b1;
            end
            else if(SDataWe) begin
                SEmpty <= 1'b0;
            end
        end
    end

endmodule

