//============================================================
//  TARGET PROJECT: *
//  TARGET SYSTEM : *
//  FUNCTION      : Uart with FIFO
//
//                      Copyright (C) 2008 by Tomoyuki Inomoto
//=============================================================
`define     FUART_CTRL      2'h0
`define     FUART_SDATA     2'h1
`define     FUART_RDATA     2'h2

//===============================================================
//   UART TOP
//===============================================================
module FifoUart(
    Clk         ,
    RstN        ,

    Clk16x115k2 ,
    Clk16x56k   ,
    Clk16x38k4  ,
    Clk16x19k2  ,
    Clk16x9k6   ,

    RxD         ,
    TxD         ,

    HostCS      ,
    HostRQ      ,
    HostRW      ,
    HostADR     ,
    HostWD      ,
    HostRD      ,

    UaSendIrq   ,
    UaRecvIrq

);
    //==================================================
    //======== port  ===================================
    //==================================================
    //======== SYSTEM ==============
    input           Clk         ;
    input           RstN        ;

    input           Clk16x115k2 ; // for uart 115200x16Hz
    input           Clk16x56k   ; // for uart  56000x16Hz
    input           Clk16x38k4  ; // for uart  38400x16Hz
    input           Clk16x19k2  ; // for uart  19200x16Hz
    input           Clk16x9k6   ; // for uart   9600x16Hz

    //======== UART ==============
    input           RxD         ;   // UART RxD
    output          TxD         ;   // UART TxD

    //======== Peripheral Bus ======
    input           HostCS      ;
    input           HostRQ      ;
    input           HostRW      ;
    input   [3:0]   HostADR     ;
    input   [31:0]  HostWD      ;
    output  [31:0]  HostRD      ;

    //======== To IrqCntr =========
    output          UaSendIrq   ; // pulse
    output          UaRecvIrq   ; // pulse

    //==================================================
    //======== wire & reg ==============================
    //==================================================
    //======== RECV DATA I/F ==============
    wire            RComp       ;   // Recv Complete
    wire            RDataRe     ;   // Recv Data Read Enable
    wire    [7:0]   RData       ;   // Recv Data

    //======== SEND DATA I/F ==============
    wire            SEmpty      ;   // Send Buf Empty
    reg             SDataWe     ;   // Send Data Write Enable
    wire    [7:0]   SData       ;   // Send Data

    //========= FIFO =======================
    wire            SFifoWe0    ;
    wire            SFifoWe     ;
    reg             SFifoWe1d   ;
    wire            SFifoEmpty  ;
    wire            SFifoFull   ;
    reg             SFifoRe     ;

    wire            RFifoWe     ;
    wire            RFifoEmpty  ;
    wire            RFifoFull   ;
    wire            RFifoAmtFull;
    wire            RFifoRe     ;
    reg             RFifoRe1d   ;
    wire    [7:0]   RFifoRD     ;

    //========= RECV BUF =======================
    reg     [10:8]  Speed       ; // 000:115.2Kbps
                                  // 001:56Kbps
                                  // 010:38.4Kbps
                                  // 011:19.2Kbps
                                  // 1**:9.6Kbps
    reg             UartCKx16   ;
    reg     [7:0]   RecvBuf     ;
    reg             RecvValid   ;
    wire            RecvBufRe   ;
    reg             RecvBufRe1d ;

    //========= IRQ =======================
    reg             SFifoEmpty1d   ;
    reg             RFifoAmtFull1d ;

    //====================================================================
    //======== BODY ======================================================
    //====================================================================
    //========= Speed =======================
    always  @(posedge Clk or negedge RstN) begin
        if(~RstN)
            Speed <= 3'h0;
        else if(HostCS & HostRQ & HostRW & (HostADR[3:2] == `FUART_CTRL))
            Speed <= HostWD[10:8];
    end
    //========= UartCKx16 =======================
    always  @(posedge Clk or negedge RstN) begin
        if(~RstN)               UartCKx16 <= 1'h0;
        else if(Speed == 3'h0)  UartCKx16 <= Clk16x115k2 ;
        else if(Speed == 3'h1)  UartCKx16 <= Clk16x56k   ;
        else if(Speed == 3'h2)  UartCKx16 <= Clk16x38k4  ;
        else if(Speed == 3'h3)  UartCKx16 <= Clk16x19k2  ;
        else                    UartCKx16 <= Clk16x9k6   ;
    end

    //===== UART ===========================
    Uart Uart(
        .Clk         ( Clk         ),
        .RstN        ( RstN        ),
        .UartCKx16   ( UartCKx16   ),

        .RxD         ( RxD         ),
        .TxD         ( TxD         ),

        .RComp       ( RComp       ),
        .RDataRe     ( RDataRe     ),
        .RData       ( RData       ),

        .SEmpty      ( SEmpty      ),
        .SDataWe     ( SDataWe     ),
        .SData       ( SData       )
    );

    //===== SEND FIFO ===========================
    assign  SFifoWe0 = (HostCS & HostRQ & HostRW & (HostADR[3:2] == `FUART_SDATA));
    always  @(posedge Clk or negedge RstN) begin
        if(~RstN) SFifoWe1d <= 1'b0;
        else      SFifoWe1d <= SFifoWe0;
    end
    assign  SFifoWe = ({SFifoWe1d, SFifoWe0} == 2'b01);
    Fifo8bx16w SFifo(
        .clk        ( Clk           ),
        .din        ( HostWD[7:0]   ),
        .rd_en      ( SFifoRe       ),
        .rst        ( ~RstN         ),
        .wr_en      ( SFifoWe       ),
        .almost_full(               ),
        .dout       ( SData         ),
        .empty      ( SFifoEmpty    ),
        .full       ( SFifoFull     )
    );

    //===== SEND CTRL ===========================
    always    @(posedge Clk or negedge RstN) begin
        if(~RstN)
            SFifoRe <= 1'b0;
        else if(~SFifoRe & ~SDataWe & ~SFifoEmpty & SEmpty)
            SFifoRe <= 1'b1;
        else
            SFifoRe <= 1'b0;
    end
    always  @(posedge Clk or negedge RstN) begin
        if(~RstN) SDataWe <= 1'b0;
        else      SDataWe <= SFifoRe;
    end

    //===== RECV FIFO ===========================
    assign  RFifoWe = RComp;
    assign  RDataRe = RComp;
    Fifo8bx16w RFifo(
        .clk        ( Clk           ),
        .din        ( RData         ),
        .rd_en      ( RFifoRe       ),
        .rst        ( ~RstN         ),
        .wr_en      ( RFifoWe       ),
        .almost_full( RFifoAmtFull  ),
        .dout       ( RFifoRD       ),
        .empty      ( RFifoEmpty    ),
        .full       ( RFifoFull     )
    );

    //===== RECV CTRL ===========================
    assign  RecvBufRe = HostCS & HostRQ & ~HostRW & (HostADR[3:2] == `FUART_RDATA);
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) RecvBufRe1d <= 1'b0;
        else      RecvBufRe1d <= RecvBufRe;
    end
    always @(posedge Clk or negedge RstN) begin
        if(~RstN)
            RecvValid <= 1'b0;
        else if({RecvBufRe1d, RecvBufRe} == 2'b01)
            RecvValid <= 1'b0;
        else if(RFifoRe)
            RecvValid <= 1'b1;
    end
    assign RFifoRe = (~RecvValid & ~RFifoEmpty);
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) RFifoRe1d <= 1'b0;
        else      RFifoRe1d <= RFifoRe;
    end
    always @(posedge Clk or negedge RstN) begin
        if(~RstN)          RecvBuf <= 8'h0;
        else if(RFifoRe1d) RecvBuf <= RFifoRD;
    end

    //===== HostRD ===========================
    assign HostRD = (HostADR[3:2] == `FUART_RDATA)? {24'h0, RecvBuf} :
                    (HostADR[3:2] == `FUART_CTRL )? {21'h0, Speed, 3'h0, SFifoFull, SFifoEmpty, RFifoFull, RFifoAmtFull, ~RecvValid} :
                                                     32'h0;

    //===== IRQ ===========================
    always @(posedge Clk or negedge RstN) begin
        if(~RstN)   SFifoEmpty1d <= 1'h0;
        else        SFifoEmpty1d <= SFifoEmpty;
    end
    assign  UaSendIrq = ({SFifoEmpty1d, SFifoEmpty} == 2'b01);

    always @(posedge Clk or negedge RstN) begin
        if(~RstN)   RFifoAmtFull1d <= 1'h0;
        else        RFifoAmtFull1d <= RFifoAmtFull;
    end
    assign  UaRecvIrq = ({RFifoAmtFull1d, RFifoAmtFull} == 2'b01);

endmodule
