//=======================================================================
//  TARGET PROJECT: *
//  TARGET SYSTEM : *
//  FUNCTION      : Debug Uart Controler
//
//                      Copyright (C) 2008 by Tomoyuki Inomoto
//=======================================================================
`define     DC_RECV_IDLE   5'h00
`define     DC_RECV_CMD    5'h01
`define     DC_RECV_D3     5'h02
`define     DC_RECV_D2     5'h03
`define     DC_RECV_D1     5'h04
`define     DC_RECV_D0     5'h05
`define     DC_RECV_ETB_W  5'h06
`define     DC_WRITE       5'h07
`define     DC_RECV_ETB_R  5'h08
`define     DC_READ        5'h09
`define     DC_SEND_SYN    5'h0A
`define     DC_SEND_D3     5'h0B
`define     DC_SEND_D2     5'h0C
`define     DC_SEND_D1     5'h0D
`define     DC_SEND_D0     5'h0E
`define     DC_SEND_ETB    5'h0F
`define     DC_RECV_BST    5'h10
`define     DC_RECV_ETB_B  5'h11

`define     SYN_CHAR    8'h16
`define     ETB_CHAR    8'h17


//===============================================================
//   DbgCtrl
//===============================================================
module DbgCtrl(
    //======== SYSTEM ==============
    Clk         ,
    RstN        ,

    //======== UART ==============
    RComp       ,
    RDataRe     ,
    RData       ,
    SEmpty      ,
    SDataWe     ,
    SData       ,

    //======== BUS ==============
    DbgWait     ,
    DbgWe       ,
    DbgRe       ,
    DbgAdr      ,
    DbgWD       ,
    DbgRD
);
    //==================================================
    //======== port  ===================================
    //==================================================
    //======== SYSTEM ==============
    input           Clk         ;
    input           RstN        ;

    //======== UART ==============
    input           RComp       ; // Recv Complete
    output          RDataRe     ; // Recv Data Read Enable
    input   [7:0]   RData       ; // Recv Data
    input           SEmpty      ; // Send Buffer Complete
    output          SDataWe     ; // Send Buffer Write Enable
    output  [7:0]   SData       ; // Send Data

    //======== DEBUG BUS ==============
    input           DbgWait     ;
    output          DbgWe       ;
    output          DbgRe       ;
    output  [3:0]   DbgAdr      ;
    output  [31:0]  DbgWD       ;
    input   [31:0]  DbgRD       ;

    //==================================================
    //======== wire & reg ==============================
    //==================================================
    reg     [4:0]   CtrlState   ;
    reg     [4:0]   CtrlState0  ;
    reg             RDataRe     ;
    reg     [31:0]  DbgWD       ;
    reg     [31:0]  RDBuf       ;
    reg     [3:0]   DbgAdr      ;
    reg             DbgWe       ;
    reg             DbgRe       ;
    reg             AdrWe       ;
    reg             D3We        ;
    reg             D2We        ;
    reg             D1We        ;
    reg             D0We        ;
    reg             SDataWe     ;
    reg     [7:0]   SData       ;
    reg     [7:0]   Burst       ;
    reg     [7:0]   Burst0      ;
    reg             BurstWe     ;
    reg     [7:0]   Parity0     ;
    reg     [7:0]   Parity      ;
    reg             ParityWe    ;

    //====================================================================
    //======== BODY ======================================================
    //====================================================================

    //======== DbgWD(WriteData)  ==========================
    always @(posedge Clk or negedge RstN) begin
        if(~RstN) begin
                         DbgAdr[3: 0]   <= 4'h0;
                         DbgWD[31: 0]   <= 32'h0;
                         RDBuf[31:0]    <= 32'h0;
                         Parity         <= 8'h0;
                         Burst          <= 8'h0;
        end
        else begin
            if(AdrWe)    DbgAdr[3: 0]   <= RData[3:0];
            if(D3We)     DbgWD[31:24]   <= RData;
            if(D2We)     DbgWD[23:16]   <= RData;
            if(D1We)     DbgWD[15: 8]   <= RData;
            if(D0We)     DbgWD[ 7: 0]   <= RData;
            if(DbgRe)    RDBuf[31:0]    <= DbgRD;
            if(ParityWe) Parity         <= Parity0;
            if(BurstWe)  Burst          <= Burst0;
        end
    end
    //======== STATE MACHINE  ==========================
    always @(posedge Clk or negedge RstN) begin
        if(~RstN)   CtrlState <= `DC_RECV_IDLE;
        else        CtrlState <= CtrlState0;
    end
    always @(CtrlState or RComp or RData or DbgWait or SEmpty or RDBuf or Parity or Burst) begin
        DbgWe       = 1'b0;
        DbgRe       = 1'b0;
        RDataRe     = 1'b0;
        AdrWe       = 1'b0;
        D3We        = 1'b0;
        D2We        = 1'b0;
        D1We        = 1'b0;
        D0We        = 1'b0;
        SDataWe     = 1'b0;
        SData       = `SYN_CHAR;
        Parity0     = 8'h0;
        ParityWe    = 1'b0;
        BurstWe     = 1'b0;
        Burst0      = Burst;
        case(CtrlState)
            `DC_RECV_IDLE  :begin
                            CtrlState0 = `DC_RECV_IDLE;
                            if(RComp) begin
                                RDataRe = 1'b1;
                                if(RData == `SYN_CHAR) begin    // RECV SYNC
                                    CtrlState0 = `DC_RECV_CMD;
                                end
                            end
                        end
            `DC_RECV_CMD   :begin
                            CtrlState0 = `DC_RECV_CMD;
                            if(RComp) begin
                                RDataRe    = 1'b1;
                                Parity0    = 8'h0;
                                ParityWe   = 1'b1;
                                if(RData[7:4] == 4'b1111) begin      // DC_WRITE COMMAND
                                    CtrlState0 = `DC_RECV_D3;
                                    AdrWe      = 1'b1;
                                    Parity0    = RData;
                                end
                                else if(RData[7:4] == 4'b0000) begin // DC_READ  COMMAND
                                    CtrlState0 = `DC_RECV_ETB_R;
                                    AdrWe      = 1'b1;
                                    Parity0    = RData;
                                    Burst0     = 8'h0;
                                    BurstWe    = 1'b1;
                                end
                                else if(RData[7:4] == 4'b0001) begin // DC_BURST_READ  COMMAND
                                    CtrlState0 = `DC_RECV_BST;
                                    AdrWe      = 1'b1;
                                    Parity0    = RData;
                                end
                                else if(RData != `SYN_CHAR) begin   // RECV SYNC
                                    CtrlState0 = `DC_RECV_IDLE;
                                    ParityWe   = 1'b0;
                                end
                            end
                        end
            `DC_RECV_D3    :begin
                            CtrlState0 = `DC_RECV_D3;
                            if(RComp) begin
                                RDataRe    = 1'b1;
                                D3We       = 1'b1;
                                CtrlState0 = `DC_RECV_D2;
                                Parity0    = RData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_RECV_D2    :begin
                            CtrlState0 = `DC_RECV_D2;
                            if(RComp) begin
                                RDataRe    = 1'b1;
                                D2We       = 1'b1;
                                CtrlState0 = `DC_RECV_D1;
                                Parity0    = RData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_RECV_D1    :begin
                            CtrlState0 = `DC_RECV_D1;
                            if(RComp) begin
                                RDataRe    = 1'b1;
                                D1We       = 1'b1;
                                CtrlState0 = `DC_RECV_D0;
                                Parity0    = RData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_RECV_D0    :begin
                            CtrlState0 = `DC_RECV_D0;
                            if(RComp) begin
                                RDataRe    = 1'b1;
                                D0We       = 1'b1;
                                CtrlState0 = `DC_RECV_ETB_W;
                                Parity0    = RData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_RECV_ETB_W :begin
                            CtrlState0 = `DC_RECV_ETB_W;
                            if(RComp) begin
                                RDataRe = 1'b1;
                                if(RData == Parity) begin    // RECV ETB
                                    CtrlState0 = `DC_WRITE;
                                end
                                else  begin                     // RECV Not ETB
                                    CtrlState0 = `DC_RECV_IDLE;
                                end
                            end
                        end
            `DC_WRITE      :begin
                            CtrlState0 = `DC_WRITE;
                            DbgWe      = 1'b1;
                            if(DbgWait == 1'b0) CtrlState0 = `DC_RECV_IDLE;
                        end
            `DC_RECV_ETB_R :begin
                            CtrlState0 = `DC_RECV_ETB_R;
                            if(RComp) begin
                                RDataRe = 1'b1;
                                if(RData == Parity) begin    // RECV ETB
                                    CtrlState0 = `DC_SEND_SYN;
                                end
                                else  begin                     // RECV Not ETB
                                    CtrlState0 = `DC_RECV_IDLE;
                                end
                            end
                        end
            `DC_READ       :begin
                            CtrlState0 = `DC_READ;
                            DbgRe      = 1'b1;
                            if(DbgWait == 1'b0) CtrlState0 = `DC_SEND_D3;
                        end
            `DC_SEND_SYN   :begin
                            CtrlState0 = `DC_SEND_SYN;
                            SData      = `SYN_CHAR;
                            if(SEmpty) begin
                                SDataWe    = 1'b1;
                                CtrlState0 = `DC_READ;
                                Parity0    = 8'h0;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_SEND_D3    :begin
                            CtrlState0 = `DC_SEND_D3;
                            SData      = RDBuf[31:24];
                            if(SEmpty) begin
                                SDataWe    = 1'b1;
                                CtrlState0 = `DC_SEND_D2;
                                Parity0    = SData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_SEND_D2    :begin
                            CtrlState0 = `DC_SEND_D2;
                            SData      = RDBuf[23:16];
                            if(SEmpty) begin
                                SDataWe    = 1'b1;
                                CtrlState0 = `DC_SEND_D1;
                                Parity0    = SData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_SEND_D1    :begin
                            CtrlState0 = `DC_SEND_D1;
                            SData      = RDBuf[15: 8];
                            if(SEmpty) begin
                                SDataWe    = 1'b1;
                                CtrlState0 = `DC_SEND_D0;
                                Parity0    = SData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_SEND_D0    :begin
                            CtrlState0 = `DC_SEND_D0;
                            SData      = RDBuf[ 7: 0];
                            if(SEmpty) begin
                                SDataWe    = 1'b1;
                                Parity0    = SData ^ Parity;
                                ParityWe   = 1'b1;
                                if(Burst == 8'h0) begin
                                    CtrlState0 = `DC_SEND_ETB;
                                end
                                else begin
                                    CtrlState0 = `DC_READ;
                                    Burst0     = Burst - 1'b1;
                                    BurstWe    = 1'b1;
                                end
                            end
                        end
            `DC_SEND_ETB   :begin
                            CtrlState0 = `DC_SEND_ETB;
                            SData      = Parity;
                            if(SEmpty) begin
                                SDataWe    = 1'b1;
                                CtrlState0 = `DC_RECV_IDLE;
                            end
                        end
            `DC_RECV_BST    :begin
                            CtrlState0 = `DC_RECV_BST;
                            if(RComp) begin
                                RDataRe    = 1'b1;
                                BurstWe    = 1'b1;
                                Burst0     = RData;
                                CtrlState0 = `DC_RECV_ETB_B;
                                Parity0    = RData ^ Parity;
                                ParityWe   = 1'b1;
                            end
                        end
            `DC_RECV_ETB_B :begin
                            CtrlState0 = `DC_RECV_ETB_B;
                            if(RComp) begin
                                RDataRe = 1'b1;
                                if(RData == Parity) begin    // RECV ETB
                                    CtrlState0 = `DC_SEND_SYN;
                                end
                                else  begin                     // RECV Not ETB
                                    CtrlState0 = `DC_RECV_IDLE;
                                end
                            end
                        end
            default     :begin
                            CtrlState0 = `DC_RECV_IDLE;
                        end
        endcase
    end
endmodule
